/*
 * _lowlevel_caches_initialization.h
 *
 *  Created on: 17 march 2020
 *      Author: Sukhanov Mikhail (suhanov@niiet.ru)
 */

// Cache way_1 (all except way_0) initialization for 1906VM016 processor (software FIX)

// ASI - Alternative Space Identifier
#define ASI_SYSTEM_CONTROL 0x02
#define CCR_ADDRESS  0x00
#define ICCR_ADDRESS 0x08
#define DCCR_ADDRESS 0x0C

#define ASI_ICACHE_DATA 0x0D
#define ASI_DCACHE_DATA 0x0F

// CCR - Cache Control Register
// CCR (ASI 0x02, address 0x00) fields:
// ...
// 15 IP - Instruction cache flush pending
// 14 DP - Data cache flush pending
// ...
// 3,2 DCS - Data cache status
// 1,0 ICS - Instruction cache status

#define DS_OFFSET  23
#define IP_OFFSET  15
#define DP_OFFSET  14
#define DCS_OFFSET  2
#define ICS_OFFSET  0

// ICS/DCS values:
#define CS_ENABLE  3
#define CS_FROZEN  1
#define CS_DISABLE 0

// ICCR - Instruction Cache Configuration Register
// DCCR - DATA Cache Configuration Register
// ICCR/DCCR (ASI 0x02, address 0x08/0x0C) fields:
// ...
// 26-24 WAYS - Number of cache ways minus 1
// 23-20 WSIZE - Size of single way (2^^WSIZE Kbytes)
// ...

#define WAYS_OFFSET  24
#define WSIZE_OFFSET 20
#define WAYS_MASK    0x7
#define WSIZE_MASK   0xF

static inline unsigned int _asi_system_control_read(unsigned int address)
  __attribute__((always_inline));
static inline unsigned int _asi_system_control_read(unsigned int address) {
	unsigned int result;
	asm volatile (
		"lda [%1] %2, %0"
		: "=r" (result)
		: "r" (address), "i" (ASI_SYSTEM_CONTROL)
	);
	return result;
}

static inline void _asi_system_control_ccr_write(unsigned int value)
  __attribute__((always_inline));
static inline void _asi_system_control_ccr_write(unsigned int value) {
	asm volatile (
		"sta %2, [%0] %1"
		:
		: "r" (CCR_ADDRESS), "i" (ASI_SYSTEM_CONTROL), "r" (value)
	);
}

static inline void _asi_system_control_ccr_write_0()
  __attribute__((always_inline));
static inline void _asi_system_control_ccr_write_0() {
	asm volatile (
		"sta %%g0, [%0] %1"
		:
		: "r" (CCR_ADDRESS), "i" (ASI_SYSTEM_CONTROL)
	);
}


static inline void _asi_dcache_data_write_0(unsigned int address)
  __attribute__((always_inline));
static inline void _asi_dcache_data_write_0(unsigned int address) {
	asm volatile (
		"sta %%g0, [%0] %1"
		:
		: "r" (address), "i" (ASI_DCACHE_DATA)
		: "memory"
	);
}

// Clear VALID fields (rewrite all tags of instruction and data caches)
static inline void _initialize_caches_tags() __attribute__((always_inline));
static inline void _initialize_caches_tags() {
	asm volatile ("flush");
}

static inline void _wait_until_flush_end() __attribute__((always_inline));
static inline void _wait_until_flush_end() {
	unsigned int cache_control_register;
	cache_control_register = _asi_system_control_read(CCR_ADDRESS);
	while (cache_control_register & ((1 << IP_OFFSET) | (1 << DP_OFFSET))) {
		cache_control_register = _asi_system_control_read(CCR_ADDRESS);
	}
}


static inline unsigned int _get_cache_way_byte_size(unsigned int cache_configuration)
  __attribute__((always_inline));
static inline unsigned int _get_cache_way_byte_size(unsigned int cache_configuration) {
	//return (1024 << ((cache_configuration >> WSIZE_OFFSET) & WSIZE_MASK));
	return (1 << ((((cache_configuration >> WSIZE_OFFSET) & WSIZE_MASK)) + 10));
}

static inline unsigned int _get_cache_byte_size(unsigned int cache_configuration)
  __attribute__((always_inline));
static inline unsigned int _get_cache_byte_size(unsigned int cache_configuration) {
	unsigned int ways;
	ways = 1 + ((cache_configuration >> WAYS_OFFSET) & WAYS_MASK);
	return (ways << ((((cache_configuration >> WSIZE_OFFSET) & WSIZE_MASK)) + 10));
}

static inline void _aligned_icache_data_write_0(
  unsigned int begin_address, unsigned int end_address
) __attribute__((always_inline));
static inline void _aligned_icache_data_write_0(
	unsigned int begin_address, unsigned int end_address
) {
	// The right way is to place this cycle into one prefetch block (8 instruction)
	// Such alignment may reduce number of external memory accesses (speed up initialization).
	unsigned int current_address;
	asm volatile (
		"mov %1, %0;"
		"  .align 32;"
		"1:;"
		"  sub %0, 4, %0;"
		"  subcc %0, %2, %%g0;" // exit from cycle before addresses of Way 0 begins
		"  bne 1b;"
		"   sta %%g0, [%0] %3;" // write to ASI 0x0D (Instruction Cache data)
		"  .align 4;"           // Restore word alignment
		: "=r" (current_address)
		: "r" (begin_address), "r" (end_address), "i" (ASI_ICACHE_DATA)
		: "cc", "memory"
	);
}

static inline void _initialize_icache_data_except_way_0() __attribute__((always_inline));
static inline void _initialize_icache_data_except_way_0() {
	unsigned int cache_configuration;
	unsigned int cache_way_byte_size;
	unsigned int cache_byte_size;

	cache_configuration = _asi_system_control_read(ICCR_ADDRESS);
	cache_way_byte_size = _get_cache_way_byte_size(cache_configuration);
	cache_byte_size = _get_cache_byte_size(cache_configuration);

	_aligned_icache_data_write_0(cache_way_byte_size, cache_byte_size);
}

static inline void _initialize_dcache_data_except_way_0() __attribute__((always_inline));
static inline void _initialize_dcache_data_except_way_0() {
	int i;
	unsigned int cache_configuration;
	unsigned int cache_way_byte_size;
	unsigned int cache_byte_size;

	cache_configuration = _asi_system_control_read(DCCR_ADDRESS);
	cache_way_byte_size = _get_cache_way_byte_size(cache_configuration);
	cache_byte_size = _get_cache_byte_size(cache_configuration);

	for (i = cache_way_byte_size; i < cache_byte_size; i += 4) {
		_asi_dcache_data_write_0(i);
	}
}


// Note 1: STA / LDA require that the processor have been in supervisor mode (S bit of %psr must be set)
// Note 2: Instruction Cache tag/data access (ASI 0x0C-0x0D) require that Instruction Cache state must be "disabled"
// Note 3: STA / LDA to ASI 0x0C-0x0F during flush lead to data_access_exception trap
static inline void _lowlevel_caches_initialization() __attribute__((always_inline));
static inline void _lowlevel_caches_initialization() {
	unsigned int value;

	// Turn off cache
	_asi_system_control_ccr_write_0();

	_initialize_caches_tags();
	_wait_until_flush_end();

	_initialize_icache_data_except_way_0();
	// Enable only instruction cache
	_asi_system_control_ccr_write(CS_ENABLE << ICS_OFFSET);

	_initialize_dcache_data_except_way_0();
	// Enable both caches and data cache snooping
	value = (1 << DS_OFFSET) | (CS_ENABLE << DCS_OFFSET) | (CS_ENABLE << ICS_OFFSET);
	_asi_system_control_ccr_write(value);
}


